/*
 Copyright (c) 2014 Shinji Tsuneyuki
 This file is distributed under the terms of the GNU General Public License version 3.
 */

#include <stdio.h>
#include <ctype.h>
#include "dtlattice.h"
#include "dtqesp.h"
#include "qtxml.h"
#include "qtexception.h"
#include "namelist.h"
#include "symbol.h"
#include <algorithm>

// Fortran̎w` 1.0d2 C̎w` 1.0e2 ɕϊ֐
static bool change_exp_format( char* ptr )
{
  bool effective = true;
  while( *ptr ){
    if( *ptr == '\'' ){
      effective = !effective;
    }

    if( effective ){
      *ptr = toupper(*ptr);
    }
    if( ptr[0] == 'D' ){
      if( ptr[1] == '+' || ptr[1] == '-' || ptr[1] == '0' ){
	ptr[0] = 'e';
      }
    }
    ptr++;
  }

  return true;
}

vector<DTQESP::AtomicSpecies::Element> DTQESP::AtomicSpecies::velement_def;

const double AU_TO_AA = 0.529177210818; // a.u. length to angstrom
const double AA_TO_AU = 1.0/AU_TO_AA;

Table DTQESP::Control::forc_conv_thr_def( 1.0e-2, 1.0e-3, 1.0e-4, 1.0e-5, 1.0e-1 );
Table DTQESP::System::ecutwfc_def( 50.0, 100.0, 200.0, 50.0, 500.0 );
Table DTQESP::Electrons::electron_maxstep_def( 50.0, 100.0, 200.0, 10.0, 500.0 );
Table DTQESP::Electrons::conv_thr_def( 1.0e-4, 1.0e-6, 1.0e-8, 1.0e-10, 1.0e-2 );
Table DTQESP::Electrons::mixing_beta_def( 0.8, 0.6, 0.4, 0.1, 0.9 );

void DTQESP::clear( void )
{
  control.calculation = "'scf'";
  control.restart_mode = "'from_scratch'";

  control.forc_conv_thr
    = Control::forc_conv_thr_def.getDefault("Normal");
  system.ecutwfc
    = System::ecutwfc_def.getDefault("Normal");
  electrons.electron_maxstep
    = (int)Electrons::electron_maxstep_def.getDefault("Normal");
  electrons.conv_thr
    = Electrons::conv_thr_def.getDefault("Normal");
  electrons.mixing_beta
    = Electrons::mixing_beta_def.getDefault("Normal");

  control.nstep  = 1;
  system.nbnd    = 0;
  system.nspin   = 1;

  kpoints.nkmesh[0] = 1;
  kpoints.nkmesh[1] = 1;
  kpoints.nkmesh[2] = 1;
  kpoints.skmesh[0] = 0;
  kpoints.skmesh[1] = 0;
  kpoints.skmesh[2] = 0;
  kpoints.nkband    = 0;

  atomicpositions.atom.clear();
}

QString DTQESP::guess( const QString& fname )
{
  FILE* fptr = fopen( fname, "rb" );
  if( fptr == NULL ){
    return "";
  }

  char buf[256];
  char svalue[32];

  bool match = false;

  while( fgets(buf,sizeof(buf),fptr) ){
    if( 1 == sscanf(buf," %s", svalue ) &&
	QString(svalue) == "&control" ){
      match = true;
      break;
    }
  }

  fclose(fptr);

  if( match ){
    return "solver qesp format";
  }

  return "";
}

bool DTQESP::load( const QString& fname )
{
  clear();

  NameListBuffer nlb(qPrintable(fname));

  try{
    char name[1024], buf[1024], value[1024], arg[32];
    int nat;

    if( !nlb.isValid() ){
      throw MyException("can not open an input file.",fname);
    }

    // Control
    if( !nlb.readList("&control") ){
      throw MyException("&control tag not found.",fname);
    }
    while( nlb.getVariable(name,value) ){
      if( false );
      else if( !strcasecmp(name,"calculation") ){
	control.calculation = QString(value);
      }
      else if( !strcasecmp(name,"restart_mode") ){
	control.restart_mode = QString(value);
      }
      else if( !strcasecmp(name,"forc_conv_thr") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &control.forc_conv_thr ) ){
	  throw MyException("broken forc_conv_thr",fname);
	}
      }
      else if( !strcasecmp(name,"nstep") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%d", &control.nstep ) ){
	  throw MyException("broken nstep",fname);
	}
      }
    }

    // System
    if( !nlb.readList("&system") ){
      throw MyException("&system tag not found.",fname);
    }
    while( nlb.getVariable(name,value) ){
      if( false );
      else if( !strcasecmp(name,"ibrav") ){
	if( 1 != sscanf(value,"%d", &system.ibrav ) ){
	  throw MyException("broken ibrav",fname);
	}
      }
      else if( !strcasecmp(name,"celldm(1)") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.celldm[1] ) ){
	  throw MyException("broken celldm(1)",fname);
	}
	system.use_celldm = true;
	system.celldm[1] *= AU_TO_AA; // converted to AA
      }
      else if( !strcasecmp(name,"celldm(2)") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.celldm[2] ) ){
	  throw MyException("broken celldm(2)",fname);
	}
	system.use_celldm = true;
      }
      else if( !strcasecmp(name,"celldm(3)") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.celldm[3] ) ){
	  throw MyException("broken celldm(3)");
	}
	system.use_celldm = true;
      }
      else if( !strcasecmp(name,"celldm(4)") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.celldm[4] ) ){
	  throw MyException("broken celldm(4)",fname);
	}
	system.use_celldm = true;
      }
      else if( !strcasecmp(name,"celldm(5)") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.celldm[5] ) ){
	  throw MyException("broken celldm(5)",fname);
	}
	system.use_celldm = true;
      }
      else if( !strcasecmp(name,"celldm(6)") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.celldm[6] ) ){
	  throw MyException("broken celldm(6)");
	}
	system.use_celldm = true;
      }
      else if( !strcasecmp(name,"A") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.A ) ){
	  throw MyException("broken A");
	}
	system.use_celldm = false;
      }
      else if( !strcasecmp(name,"B") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.B ) ){
	  throw MyException("broken B");
	}
	system.use_celldm = false;
      }
      else if( !strcasecmp(name,"C") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.C ) ){
	  throw MyException("broken C");
	}
	system.use_celldm = false;
      }
      else if( !strcasecmp(name,"cosAB") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.cosAB ) ){
	  throw MyException("broken cosAB");
	}
	system.use_celldm = false;
      }
      else if( !strcasecmp(name,"cosAC") || !strcasecmp(name,"cosCA") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.cosCA ) ){
	  throw MyException("broken cosCA");
	}
	system.use_celldm = false;
      }
      else if( !strcasecmp(name,"cosBC") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.cosBC ) ){
	  throw MyException("broken cosBC");
	}
	system.use_celldm = false;
      }
      else if( !strcasecmp(name,"nat") ){
	if( 1 != sscanf(value,"%d", &nat ) ){
	  throw MyException("broken nat");
	}
      }
      else if( !strcasecmp(name,"ecutwfc") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &system.ecutwfc ) ){
	  throw MyException("broken ecutwfc");
	}
      }
      else if( !strcasecmp(name,"nbnd") ){
	if( 1 != sscanf(value,"%d", &system.nbnd ) ){
	  throw MyException("broken nbnd");
	}
      }
      else if( !strcasecmp(name,"nspin") ){
	if( 1 != sscanf(value,"%d", &system.nspin ) ){
	  throw MyException("broken nspin");
	}
      }
    }

    // Electrons
    if( !nlb.readList("&electrons") ){
      throw MyException("&electrons tag not found.",fname);
    }
    while( nlb.getVariable(name,value) ){
      if( false );
      else if( !strcasecmp(name,"electron_maxstep") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%d", &electrons.electron_maxstep ) ){
	  throw MyException("broken electron_maxstep");
	}
      }
      else if( !strcasecmp(name,"conv_thr") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &electrons.conv_thr ) ){
	  throw MyException("broken conv_thr");
	}
      }
      else if( !strcasecmp(name,"mixing_beta") ){
	change_exp_format(value);
	if( 1 != sscanf(value,"%lf", &electrons.mixing_beta ) ){
	  throw MyException("broken mixing_beta");
	}
      }
    }

    // ATOMIC_POSITIONS
    if( !nlb.warpLine("ATOMIC_POSITIONS",arg) ){
      throw MyException("ATOMIC_POSITIONS section not found.",fname);
    }
    if( false );
    else if( strcasecmp(arg,"") == 0 ||
	     strcasecmp(arg,"alat") == 0 ||
	     strcasecmp(arg,"(alat)") == 0 ){
      atomicpositions.unit = 0;
    }
    else if( strcasecmp(arg,"bohr") == 0 ||
	     strcasecmp(arg,"(bohr)") == 0){
      atomicpositions.unit = 1;
    }
    else if( strcasecmp(arg,"angstrom") == 0 ||
	     strcasecmp(arg,"(angstrom)") == 0 ){
      atomicpositions.unit = 2;
    }
    else if( strcasecmp(arg,"crystal") == 0 ||
	     strcasecmp(arg,"(crystal)") == 0 ){
      atomicpositions.unit = 3;
    }
    atomicpositions.atom.resize( nat );
    for( int i=0; i<nat; i++ ){
      char elem[32];
      Position pos;

      if( nlb.readExtra(buf) &&
	  4 == sscanf(buf,"%s %lf %lf %lf", elem, &pos.x, &pos.y, &pos.z ) ){
	atomicpositions.atom[i].name = QString(elem);
	atomicpositions.atom[i].number
	  = ElementSymbol::getAtomicNumber( atomicpositions.atom[i].name );
	atomicpositions.atom[i].coords = pos;
      }
      else{
	throw MyException("broken ATOMIC_POSITIONS data");
      }
    }

    // K_POINTS
    if( !nlb.warpLine("K_POINTS",arg) ){
      throw MyException("K_POINTS section not found.",fname);
    }
    else{
      if( strcasecmp(arg,"gamma") == 0 ){
      }
      else if( strcasecmp(arg,"automatic") == 0 ){
	if( nlb.readExtra(buf) &&
	    6 == sscanf(buf," %d %d %d %d %d %d",
			&kpoints.nkmesh[0], &kpoints.nkmesh[1], &kpoints.nkmesh[2],
			&kpoints.skmesh[0], &kpoints.skmesh[1], &kpoints.skmesh[2] ) ){
	}
	else{
	  throw MyException("broken K_POINTS data");
	}
      }
      else if( strcmp(arg,"crystal") == 0 ||
	       strcmp(arg,"(crystal)") == 0 ){
	if( nlb.readExtra(buf) &&
	    1 == sscanf(buf," %d", &kpoints.nkband ) ){
	}
	else{
	  throw MyException("broken K_POINTS data");
	}
      }
    }

    if( system.ibrav == 0 ){
      if( !nlb.warpLine("CELL_PARAMETERS",arg) ){
	throw MyException("CELL_PARAMETERS section not found.",fname);
      }

      if( nlb.readExtra(buf) &&
	  3 == sscanf(buf," %lf %lf %lf",
		      &cellparameters.La.x, &cellparameters.La.y, &cellparameters.La.z ) ){
      }
      else{
	throw MyException("broken CELL_PARAMETERS data");
      }

      if( nlb.readExtra(buf) &&
	  3 == sscanf(buf," %lf %lf %lf",
		      &cellparameters.Lb.x, &cellparameters.Lb.y, &cellparameters.Lb.z ) ){
      }
      else{
	throw MyException("broken CELL_PARAMETERS data");
      }

      if( nlb.readExtra(buf) &&
	  3 == sscanf(buf," %lf %lf %lf",
		      &cellparameters.Lc.x, &cellparameters.Lc.y, &cellparameters.Lc.z ) ){
      }
      else{
	throw MyException("broken CELL_PARAMETERS data");
      }

      if( strcasecmp(arg,"bohr") == 0 ||
	  strcasecmp(arg,"(bohr)") == 0){
	cellparameters.La *= AU_TO_AA;
	cellparameters.Lb *= AU_TO_AA;
	cellparameters.Lc *= AU_TO_AA;
      }
      else if( strcasecmp(arg,"angstrom") == 0 ||
	       strcasecmp(arg,"(angstrom)") == 0 ){
      }
      else{
	cellparameters.La *= system.celldm[1];
	cellparameters.Lb *= system.celldm[1];
	cellparameters.Lc *= system.celldm[1];
      }
    }

    convertCell();
    convertAtom();
  }
  catch( const MyException& e ){
    MyException::critical(e);
    return false;
  }

  return true;
}

bool DTQESP::save( const DTLattice& lattice, const QString& fname ) const
{
  FILE* fptr = fopen( fname, "wb" );

  try{
    if( fptr == NULL ){
      throw MyException("can not create a QESP file.",fname );
    }

    vector<QString> velem;
    for( int i=0; i<(int)atomicpositions.atom.size(); i++ ){
      const QString& elem = atomicpositions.atom[i].name;
      if( std::find( velem.begin(), velem.end(), elem ) == velem.end() ){
	velem.push_back(elem);
      }
    }

    // Control
    {
      fprintf(fptr, "&control\n");

      if( false );
      else if( control.calculation == "'charge'" ||
	       control.calculation == "'scf'" ){
	fprintf(fptr, "  calculation   = 'scf'\n");
      }
      else if( control.calculation == "'bandmap'" ){
	fprintf(fptr, "  calculation   = 'bands'\n");
      }
      else if( control.calculation == "'dosmap'" ){
	fprintf(fptr, "  calculation   = 'nscf'\n");
      }
      else if( control.calculation == "'relax'" ){
	fprintf(fptr, "  calculation   = 'vc-relax'\n");
      }
      else{
	// ???
      }
      fprintf(fptr, "  restart_mode  = %s\n", qPrintable(control.restart_mode) );
      fprintf(fptr, "  pseudo_dir = './'\n");
      fprintf(fptr, "  outdir     = './'\n");
      fprintf(fptr, "  forc_conv_thr = %e\n", control.forc_conv_thr );
      fprintf(fptr, "  etot_conv_thr = %e\n", control.forc_conv_thr*0.1 );
      fprintf(fptr, "  nstep         = %d\n", control.nstep );
      fprintf(fptr, "/\n");
    }

    // System
    {
      fprintf(fptr, "&system\n");
      fprintf(fptr, "  ibrav   = 0\n");
      fprintf(fptr, "  nat     = %d\n", (int)atomicpositions.atom.size() );
      fprintf(fptr, "  ntyp    = %d\n", (int)velem.size() );
      fprintf(fptr, "  ecutwfc = %f\n", system.ecutwfc );
      fprintf(fptr, "  ecutrho = %f\n", system.ecutwfc*4.0 );
      fprintf(fptr, "  nbnd    = %d\n", system.nbnd );
      fprintf(fptr, "  nspin   = %d\n", system.nspin );
      if( control.calculation == "'charge'" ||
	  control.calculation == "'scf'" || 
	  control.calculation == "'bandmap'" ){
	fprintf(fptr, "  occupations = 'smearing'\n");
	fprintf(fptr, "  smearing = 'mv'\n");
	fprintf(fptr, "  degauss = 0.04\n");
      }
      if( control.calculation == "'relax'" ){
	fprintf(fptr, "  occupations = 'smearing'\n");
	fprintf(fptr, "  smearing = 'gaussian'\n");
	fprintf(fptr, "  degauss = 0.04\n");
      }
      else if( control.calculation == "'dosmap'" ){
	fprintf(fptr, "  occupations = 'tetrahedra'\n");
      }
      if( control.calculation == "'dosmap'" ){
	fprintf(fptr, "  nosym = .true.\n");
	fprintf(fptr, "  noinv = .true.\n");
      }
      fprintf(fptr, "/\n");
    }

    // Electrons
    {
      fprintf(fptr, "&electrons\n");
      fprintf(fptr, "  electron_maxstep = %d\n", electrons.electron_maxstep );
      fprintf(fptr, "  conv_thr = %e\n", electrons.conv_thr );
      fprintf(fptr, "  mixing_beta = %f\n", electrons.mixing_beta );
      fprintf(fptr, "/\n");

      if( control.calculation == "'relax'" ){
	fprintf(fptr, "&ions\n");
	fprintf(fptr, "  ion_dynamics='bfgs'\n");
	fprintf(fptr, "/\n");
	fprintf(fptr, "&cell\n");
	fprintf(fptr, "/\n");
      }
    }

    // Dos
    if( control.calculation == "'dosmap'" ){
      fprintf(fptr, "&dos\n");
      fprintf(fptr, "  Emax   = %f\n", +20.0 );
      fprintf(fptr, "  Emin   = %f\n", -20.0 );
      fprintf(fptr, "  DeltaE = %f\n",   0.1 );
      fprintf(fptr, "/\n");
    }

    // Bands
    if( control.calculation == "'bandmap'" ){
      fprintf(fptr, "&bands\n");
      fprintf(fptr, "  no_overlap = .true.\n");
      fprintf(fptr, "  lsym  = .true.\n");
      fprintf(fptr, "/\n");
    }

    // ATOMIC_SPECIES
    {
      fprintf( fptr, "ATOMIC_SPECIES\n" );

      for( int i=0; i<(int)velem.size(); i++ ){
	const int number = ElementSymbol::getAtomicNumber( velem[i] );

	fprintf( fptr, " %2s %8.4f %s\n",
		 qPrintable(velem[i]),
		 AtomicSpecies::velement_def[number].mass,
		 qPrintable(AtomicSpecies::velement_def[number].psfile) );
      }
    }

    // ATOMIC_POSITIONS
    {
      fprintf( fptr, "ATOMIC_POSITIONS crystal\n" );
      for( int i=0; i<(int)atomicpositions.atom.size(); i++ ){
	fprintf( fptr, " %2s %.9f %.9f %.9f\n",
		 qPrintable(atomicpositions.atom[i].name),
		 atomicpositions.atom[i].coords.x,
		 atomicpositions.atom[i].coords.y,
		 atomicpositions.atom[i].coords.z );
      }
    }

    // K_POINTS
    {
      if( control.calculation == "'bandmap'" ){
	const vector<BrillouinPoint> vsymmK 
	  = lattice.cell.vsymmK;

	fprintf( fptr, "K_POINTS crystal\n" );
	fprintf(fptr, " %d\n", (int)vsymmK.size() );
	for( int i=0; i<(int)vsymmK.size(); i++ ){
	  fprintf( fptr, " %+f %+f %+f 1.0 %s\n",
		   vsymmK[i].position.x,
		   vsymmK[i].position.y,
		   vsymmK[i].position.z,
		   qPrintable(vsymmK[i].label) );
	}
      }
      else{
	fprintf( fptr, "K_POINTS automatic\n" );
	fprintf( fptr, " %d %d %d %d %d %d\n",
		 kpoints.nkmesh[0], kpoints.nkmesh[1], kpoints.nkmesh[2],
		 kpoints.skmesh[0], kpoints.skmesh[1], kpoints.skmesh[2] );
      }
    }

    // CELL_PARAMETERS
    {
      fprintf( fptr, "CELL_PARAMETERS angstrom\n" );

      fprintf( fptr," %+.9f %+.9f %+.9f\n",
	       cellparameters.La.x, cellparameters.La.y, cellparameters.La.z );
      fprintf( fptr," %+.9f %+.9f %+.9f\n",
	       cellparameters.Lb.x, cellparameters.Lb.y, cellparameters.Lb.z );
      fprintf( fptr," %+.9f %+.9f %+.9f\n",
	       cellparameters.Lc.x, cellparameters.Lc.y, cellparameters.Lc.z );
    }

    fclose(fptr);
  }
  catch( const MyException& e ){
    MyException::critical(e);
    if(fptr) fclose(fptr);
    return false;
  }
  return true;
}



void DTQESP::convertCell( void )
{
  /*
    memo
    CELL_PARAMETERS is given in AA or AU. converted to AA.
    system.celldm[1] is converted to AA.
    system.A is given in AA.
  */

  switch( system.ibrav ){

  case 0 : { // free
    // use CELL_PARAMETERS
    cellparameters.L  = 1.0;
    cellparameters.Ea = cellparameters.La;
    cellparameters.Eb = cellparameters.Lb;
    cellparameters.Ec = cellparameters.Lc;
  } break;

  case 1 : { // cubic simple
    double a;
    if( system.use_celldm ){
      a = system.celldm[1];
      system.A = a;
    }
    else{
      a = system.A;
    }
    cellparameters.L  = a;
    cellparameters.Ea = Position(1.0,0.0,0.0);
    cellparameters.Eb = Position(0.0,1.0,0.0);
    cellparameters.Ec = Position(0.0,0.0,1.0);
  } break;

  case 2 : { // cubic face
    double a;
    if( system.use_celldm ){
      a = system.celldm[1]*0.5;
    }
    else{
      a = system.A*M_SQRT2*0.5;
    }
    cellparameters.L  = a;
    cellparameters.Ea = Position(-1.0,0.0,+1.0);
    cellparameters.Eb = Position(0.0,+1.0,+1.0);
    cellparameters.Ec = Position(-1.0,+1.0,0.0);
  } break;

  case 3 : { // cubic body
    double a;
    if( system.use_celldm ){
      a = system.celldm[1]*0.5;
    }
    else{
      a = system.A*M_SQRT3/3.0;
    }
    cellparameters.L  = a;
    cellparameters.Ea = Position(+1.0,+1.0,+1.0);
    cellparameters.Eb = Position(-1.0,+1.0,+1.0);
    cellparameters.Ec = Position(-1.0,-1.0,+1.0);
  } break;

  case 4 : { // hexagonal
    double a, ca;
    if( system.use_celldm ){
      a  = system.celldm[1];
      ca = system.celldm[3];
    }
    else{
      a  = system.A;
      ca = system.C/system.A;
    }
    cellparameters.L  = a;
    cellparameters.Ea = Position(+1.0, 0.0, 0.0);
    cellparameters.Eb = Position(-0.5,+M_SQRT3*0.5, 0.0);
    cellparameters.Ec = Position( 0.0, 0.0, ca );
  } break;

  case 5 : { // trigonal, its 3fold axis is c
    double a, cosBC;
    if( system.use_celldm ){
      a = system.celldm[1];
      cosBC = system.celldm[4];
    }
    else{
      a = system.A;
      cosBC = system.cosBC;
      /*
	|Ea|^2 = (tx*tx + ty*ty + tz*tz)
	= (1.0-cosBC)/2 + (1.0-cosBC)/6 + (1.0+2*cosBC)/3
	= (3.0-3*cosBC)/6 + (1.0-cosBC)/6 + (2.0+4*cosBC)/6
	= 1.0
      */
    }
    const double tx = sqrt((1.0-cosBC)/2);
    const double ty = sqrt((1.0-cosBC)/6);
    const double tz = sqrt((1.0+2*cosBC)/3);
    cellparameters.L  = a;
    cellparameters.Ea = Position(+tx,-ty,tz);
    cellparameters.Eb = Position(0.0,2.0*ty,tz);
    cellparameters.Ec = Position(-tx,-ty,tz);
  } break;

  case -5 : { // trigonal, its 3fold axis is <111>
    double a, cosBC;
    if( system.use_celldm ){
      a = system.celldm[1]*M_SQRT3/3.0;
      cosBC = system.celldm[4];
      /*
	|Ea|^2 = u*u + 2*v*v
	= (tz - 2.0*M_SQRT2*ty)^2 + 2*(tz +     M_SQRT2*ty)^2
	= tz*tz - 4.0*M_SQRT2*ty*tz + 8.0*ty*ty
	+ 2*tz*tz + 4.0*M_SQRT2*ty*tz + 4.0*ty*ty
	= 3*tz*tz + 12.0*ty*ty
	= 3*(1.0+2*cosBC)/3.0 + 12*(1.0-cosBC)/6.0
	= 1.0+2*cosBC + 2*(1.0-cosBC) = 3
      */
    }
    else{
      a = system.A*M_SQRT3/3.0;
      cosBC = system.cosBC;
    }
    const double ty = sqrt((1.0-cosBC)/6.0);
    const double tz = sqrt((1.0+2*cosBC)/3.0);
    const double u  = tz - 2.0*M_SQRT2*ty;
    const double v  = tz +     M_SQRT2*ty;
    cellparameters.L  = a;
    cellparameters.Ea = Position(u,v,v);
    cellparameters.Eb = Position(v,u,v);
    cellparameters.Ec = Position(v,v,u);
  } break;

  case 6 : { // tetragonal simple
    double a, ca;
    if( system.use_celldm ){
      a  = system.celldm[1];
      ca = system.celldm[3];
    }
    else{
      a  = system.A;
      ca = system.C/system.A;
    }
    cellparameters.L  = a;
    cellparameters.Ea = Position(1.0,0.0,0.0);
    cellparameters.Eb = Position(0.0,1.0,0.0);
    cellparameters.Ec = Position(0.0,0.0,ca );
  } break;

  case 7 : { // tetragonal body
    double a, ca;
    if( system.use_celldm ){
      a  = system.celldm[1];
      ca = system.celldm[3];
    }
    else{
      const double cosAB = system.cosAB;
      const double cosBC = system.cosBC;
      const double cosCA = system.cosCA;
      if( cosAB != cosCA ){
	throw MyException("Error: tetragonal with cosAB != cosCA");
      }
      if( cosAB == cosBC ){
	throw MyException("Error: tetragonal with cosAB == cosBC");
      }
      a  = system.A*sqrt(2.0-2.0*cosAB);
      ca = sqrt(2.0*cosAB/(1.0-cosAB));
      /*
	v1 = (a,-a,c)/2
	v2 = (a,+a,c)/2

	cosAB = v1.v2/|v1||v2|
	= (a*a-a*a+c*c)/4 / (a*a+a*a+c*c)/4
	= c*c/(2*a*a+c*c) = ca*ca/(2+ca*ca)
	(2+ca*ca)*cosAB = ca*ca
	2*cosAB = ca*ca*(1-cosAB)
	ca*ca = 2*cosAB/(1-cosAB)

	A^2 = |v1|^2 = (2*a*a+c*c)/4 = (2+ca*ca)/4 *a*a
	  
	a*a = 4*A^2/(2+ca*ca)
	a = A*2/sqrt(2+ca*ca)
	= A*2/sqrt(2+2*cosAB/(1-cosAB))
	= A*2/sqrt(2/(1-cosAB))
	= A*sqrt(2-2*cosAB)
      */
    }
    cellparameters.L  = a*0.5;
    cellparameters.Ea = Position(+1.0,-1.0,ca );
    cellparameters.Eb = Position(+1.0,+1.0,ca );
    cellparameters.Ec = Position(-1.0,-1.0,ca );
  } break;

  case 8 : { // orthorhombic simple
    double a, b, c;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
    }
    else{
      a = system.A;
      b = system.B;
      c = system.C;
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a  ,0.0,0.0);
    cellparameters.Eb = Position(0.0,b,  0.0);
    cellparameters.Ec = Position(0.0,0.0,c  );
  } break;

  case 9 : { // orthorhombic base
    double a, b, c;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
    }
    else{
      a = system.A;
      c = system.C;
      b = system.A*sqrt((1.0+system.cosAB)/(1.0-system.cosAB));
      /*
	v1 = (+a,b,0)/2
	v2 = (-a,b,0)/2
	cosAB = v1.v2/|v1||v2| = (-a*a+b*b)/(a*a+b*b)
	= (-1+(b/a)^2) / (1+(b/a)^2)
	(1+(b/a)^2)*cosAB = -1 + (b/a)^2
	(b/a)^2 *(1-cosAB) = 1+cosAB
	(b/a)^2 = (1+cosAB)/(1-cosAB)
      */
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(+a*0.5,b*0.5,0.0);
    cellparameters.Eb = Position(-a*0.5,b*0.5,0.0);
    cellparameters.Ec = Position(0.0,0.0,c);
  } break;

  case -9 : { // orthorhombic base, alternate
    double a, b, c;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
    }
    else{
      a = system.A;
      c = system.C;
      b = system.A*sqrt((1.0-system.cosAB)/(1.0+system.cosAB));
      /*
	v1 = (a,-b,0)/2
	v2 = (a,+b,0)/2
	cosAB = v1.v2/|v1||v2| = (a*a-b*b)/(a*a+b*b)
	= (1-(b/a)^2) / (1+(b/a)^2)
	(1+(b/a)^2)*cosAB = 1 - (b/a)^2
	(b/a)^2*(1+cosAB) = 1 - cosAB
	(b/a)^2 = (1-cosAB)/(1+cosAB)
      */
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a*0.5,-b*0.5,0.0);
    cellparameters.Eb = Position(a*0.5,+b*0.5,0.0);//QE doc is wrong
    cellparameters.Ec = Position(0.0,0.0,c);
  } break;

  case 10 : { // orthorhombic face
    double a, b, c;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
    }
    else{
      const double A = system.A;
      const double B = system.B;
      const double C = system.C;
      a = sqrt(+2.0*A*A+2.0*B*B-2.0*C*C);
      b = sqrt(-2.0*A*A+2.0*B*B+2.0*C*C);
      c = sqrt(+2.0*A*A-2.0*B*B+2.0*C*C);
      /*
	v1 = (a,0,c)/2
	v2 = (a,b,0)/2
	v3 = (0,b,c)/2
	A*A = |v1|^2 = (a*a + c*c)/4
	B*B = |v2|^2 = (a*a + b*b)/4
	C*C = |v3|^2 = (b*b + c*c)/4
	A*A + B*B + C*C = (a*a + b*b + c*c)/2

	2*A*A + 2*B*B + 2*C*C = a*a + b*b + c*c
	4*A*A = a*a + c*c
	4*B*B = a*a + b*b
	4*C*C = b*b + c*c
	a*a = +2*A*A + 2*B*B - 2*C*C
	b*b = -2*A*A + 2*B*B + 2*C*C
	c*c = +2*A*A - 2*B*B + 2*C*C
      */
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a*0.5,0.0,c*0.5);
    cellparameters.Eb = Position(a*0.5,b*0.5,0.0);
    cellparameters.Ec = Position(0.0,b*0.5,c*0.5);
  } break;

  case 11 : { // orthorhombic body
    double a, b, c;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
    }
    else{
      const double A = system.A;
      const double cosAB = system.cosAB;
      const double cosBC = system.cosBC;
      const double cosCA = system.cosCA;
      a = M_SQRT2*sqrt(cosBC-cosCA)*A;
      b = M_SQRT2*sqrt(cosAB-cosCA)*A;
      c = M_SQRT2*sqrt(cosAB+cosBC)*A;
      /*
	v1=(+a,+b,c)/2
	v2=(-a,+b,c)/2
	v3=(-a,-b,c)/2
	A*A = |v1|^2 = (a*a + b*b + c*c)/4
	a*a + b*b + c*c = 4*A*A

	cosAB = v1.v2/|v1||v2| = (- a*a + b*b + c*c)/(a*a + b*b + c*c)
	4*A*A*cosAB = - a*a + b*b + c*c
	cosBC = v2.v3/|v2||v3| = (+ a*a - b*b + c*c)/(a*a + b*b + c*c)
	4*A*A*cosBC = + a*a - b*b + c*c
	cosCA = v3.v1/|v3||v1| = (- a*a - b*b + c*c)/(a*a + b*b + c*c)
	4*A*A*cosCA = - a*a - b*b + c*c

	2*A*A*(cosAB+cosBC) = c*c
	2*A*A*(cosBC-cosCA) = a*a
	2*A*A*(cosAB-cosCA) = b*b
      */
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(+a*0.5,+b*0.5,c*0.5);
    cellparameters.Eb = Position(-a*0.5,+b*0.5,c*0.5);
    cellparameters.Ec = Position(-a*0.5,-b*0.5,c*0.5);
  } break;

  case 12 : { // monoclinic simple unique axis is c
    double a, b, c;
    double cosAB, sinAB;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
      cosAB = system.celldm[4];
      sinAB = sqrt(1.0-cosAB*cosAB);
    }
    else{
      a = system.A;
      b = system.B;
      c = system.C;
      cosAB = system.cosAB;
      sinAB = sqrt(1.0-cosAB*cosAB);
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a,0.0,0.0);
    cellparameters.Eb = Position(b*cosAB,b*sinAB,0.0);
    cellparameters.Ec = Position(0.0,0.0,c);
  } break;

  case -12 : { // monoclinic simple unique axis is b
    double a, b, c;
    double cosCA, sinCA;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
      cosCA = system.celldm[5];
      sinCA = sqrt(1.0-cosCA*cosCA);
    }
    else{
      a = system.A;
      b = system.B;
      c = system.C;
      cosCA = system.cosCA;
      sinCA = sqrt(1.0-cosCA*cosCA);
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a,0.0,0.0);
    cellparameters.Eb = Position(0.0,b,0.0);
    cellparameters.Ec = Position(c*cosCA,0.0,c*sinCA);
  } break;

  case 13 : { // monoclinic base
    double a, b, c;
    double cosAB, sinAB;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
      cosAB = system.celldm[4];
      sinAB = sqrt(1.0-cosAB*cosAB);
    }
    else{
      a = system.A;
      b = system.B;
      c = system.C;
      cosAB = system.cosAB;
      sinAB = sqrt(1.0-cosAB*cosAB);
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a*0.5,0.0,-c*0.5);
    cellparameters.Eb = Position(b*cosAB,b*sinAB,0.0);
    cellparameters.Ec = Position(a*0.5,0.0,+c*0.5);
  } break;

  case 14 : { // triclinic
    double a, b, c;
    double cosAB, sinAB;
    double cosBC;
    double cosCA;
    if( system.use_celldm ){
      a = system.celldm[1];
      b = system.celldm[2]*a;
      c = system.celldm[3]*a;
      cosBC = system.celldm[4];
      cosCA = system.celldm[5];
      cosAB = system.celldm[6];
      sinAB = sqrt(1.0-cosAB*cosAB);
    }
    else{
      a = system.A;
      b = system.B;
      c = system.C;
      cosBC = system.cosBC;
      cosCA = system.cosCA;
      cosAB = system.cosAB;
      sinAB = sqrt(1.0-cosAB*cosAB);
    }
    cellparameters.L  = 1.0;
    cellparameters.Ea = Position(a,0.0,0.0);
    cellparameters.Eb = Position(b*cosAB,b*sinAB,0.0);
    cellparameters.Ec =
      Position(c*cosCA,
	       c*(cosBC-cosCA*cosAB)/sinAB,
	       c*sqrt(1.0+2.0*cosBC*cosCA*cosAB
		      - cosBC*cosBC-cosCA*cosCA-cosAB*cosAB)/sinAB );
  } break;
  }

  cellparameters.La = cellparameters.Ea*cellparameters.L;
  cellparameters.Lb = cellparameters.Eb*cellparameters.L;
  cellparameters.Lc = cellparameters.Ec*cellparameters.L;
}


void DTQESP::convertAtom( void )
{
  // convert atoms in units of some kinds to cartesian coordinates in Angstrom.
  for( int i=0; i<(int)atomicpositions.atom.size(); i++ ){
    Position& r = atomicpositions.atom[i].coords;

    switch( atomicpositions.unit ){
    case 0 : { // alat
      // atomic positions are in cartesian coordinates,
      // in units of the lattice parameter.
      if( system.use_celldm ){
	r = r * system.celldm[1];
      }
      else{
	r = r * system.A;
      }
    } break;
    case 1 : { // bohr
      // atomic positions are in cartesian coordinate,
      // in atomic units.
      r = r * AU_TO_AA;
    } break;
    case 2 : { // angstrom
      // atomic positions are in cartesian coordinate,
      // in Angstrom.
      // r = r; // no conversion
    } break;

    case 3 : { // crystal
      // atomic positions are in crystal coordinates,
      // in relative coordinates of the primitive lattice vectors
      r =
	+ r.a * cellparameters.La
	+ r.b * cellparameters.Lb
	+ r.c * cellparameters.Lc;
    } break;
    }
  }

  const Position& La = cellparameters.La;
  const Position& Lb = cellparameters.Lb;
  const Position& Lc = cellparameters.Lc;
  const double V     = La*(Lb%Lc);
  const Position  Ka = (Lb%Lc)*(1.0/V);
  const Position  Kb = (Lc%La)*(1.0/V);
  const Position  Kc = (La%Lb)*(1.0/V);

  // convert atoms in cartesian coordinates in Angstrom
  // to relative coordinates of the primitive lattice vectors
  for( int i=0; i<(int)atomicpositions.atom.size(); i++ ){
    Position& r = atomicpositions.atom[i].coords;
    r = Coordinates( r*Ka, r*Kb, r*Kc );
  }

}
