/*
 Copyright (c) 2014 Shinji Tsuneyuki
 This file is distributed under the terms of the GNU General Public License version 3.
 */

/*!
 \file glrecplattice.cc
 \brief tiqGL\̃NX
*/

#include "glrecplattice.h"
#include "dtmodel.h"
#include "qtmisc.h"
#include "qtexception.h"

GLRecpLattice::GLRecpLattice( DTModel& _model ) : model(_model)
{
  setFocusPolicy( Qt::ClickFocus );
  setMinimumSize( 512, 512 );
  setSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding );

  connect( &model, SIGNAL(changed()),   this, SLOT(update()) );
  connect(   this, SIGNAL(changed()), &model, SLOT(update()) );

  popupmenu = new QMenu(this);
  popupmenu->addSeparator();
  popupmenu->addAction( new QAction(tr("reset to xy-plane"),this) );
  popupmenu->addAction( new QAction(tr("reset to yz-plane"),this) );
  popupmenu->addAction( new QAction(tr("reset to zx-plane"),this) );
  popupmenu->addSeparator();
  popupmenu->addAction( new QAction(tr("change perspective/orthogonal"),this) );
  popupmenu->addSeparator();
  popupmenu->addAction( new QAction(tr("save image"),this) );
  connect(popupmenu, SIGNAL(triggered(QAction*)), this, SLOT(menuEvent(QAction*)));

  perspective = false;
}


void GLRecpLattice::initializeGL( void )
{
  glClearColor( model.gloption.window.background );

  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity();
  gluLookAt( model.gloption.location.eye,
	     model.gloption.location.gaze,
	     model.gloption.location.up );

  glEnable( GL_DEPTH_TEST );
  glEnable( GL_NORMALIZE );

  glShadeModel( GL_SMOOTH );
  glLightModeli(  GL_LIGHT_MODEL_TWO_SIDE, 1 );
  glLightModeli(  GL_LIGHT_MODEL_LOCAL_VIEWER, 1 );
  glLightModeldv( GL_LIGHT_MODEL_AMBIENT, model.gloption.light.ambient );
  glEnable( GL_LIGHTING );

  glLightdv( GL_LIGHT0, GL_POSITION, model.gloption.light.direction );
  glLightdv( GL_LIGHT0, GL_DIFFUSE,  model.gloption.light.diffuse );
  glLightdv( GL_LIGHT0, GL_SPECULAR, model.gloption.light.specular );
  glEnable( GL_LIGHT0 );

  glLightdv( GL_LIGHT1, GL_POSITION, model.gloption.light.second_direction );

  glLightdv( GL_LIGHT1, GL_DIFFUSE,  model.gloption.light.diffuse );
  glLightdv( GL_LIGHT1, GL_SPECULAR, model.gloption.light.specular );
  if( model.gloption.light.second ){
    glEnable( GL_LIGHT1 );
  }

  glFogf( GL_FOG_MODE, GL_LINEAR );
  glFogf( GL_FOG_START,   model.gloption.location.world );
  glFogf( GL_FOG_END, 1.6*model.gloption.location.world );
  glFogfv( GL_FOG_COLOR,  model.gloption.window.background );
  if( model.gloption.light.fog ){
    glEnable( GL_FOG );
  }

  glMaterialdv( GL_FRONT_AND_BACK, GL_SPECULAR,  model.gloption.light.specular  );
  glMateriald ( GL_FRONT_AND_BACK, GL_SHININESS, model.gloption.light.shininess );

  glClearColor( model.gloption.window.background );
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}


void GLRecpLattice::gluProjection( int width, int height, bool perspective )
{
  if( perspective ){
    const double aspect = (double)width/height;
    gluPerspective( model.gloption.location.pangle,
		    aspect,
		    model.gloption.location.pnear,
		    model.gloption.location.pfar );
  }else{
    const double t = tan(model.gloption.location.pangle*0.5*M_PI/180.0);
    const double aspect = (double)width/height;
    const double top    = 1.000*sqrt(model.gloption.location.pnear*model.gloption.location.pfar) * t;
    const double bottom = -top;
    const double right  = top*aspect;
    const double left   = -right;

    glOrtho( left, right, bottom, top, 
	     model.gloption.location.pnear,
	     model.gloption.location.pfar );
  }
}

void GLRecpLattice::resizeGL(int width, int height)
{
  glViewport( 0, 0, width, height );

  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();

  gluProjection( width, height, perspective );

  glMatrixMode( GL_MODELVIEW );
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
}

void GLRecpLattice::paintGL( void )
{
  if( model.gloption.light.second ){
    glEnable( GL_LIGHT1 );
  }else{
    glDisable( GL_LIGHT1 );
  }
  if( model.gloption.light.fog ){
    glFogfv( GL_FOG_COLOR, model.gloption.window.background );
    glEnable( GL_FOG );
  }else{
    glDisable( GL_FOG );
  }
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

  draw();
}

// mouse drag begin
void GLRecpLattice::mousePressEvent( QMouseEvent* ev )
{
  if( !(ev->buttons()&Qt::LeftButton) ) return;

  int x = ev->x(), y = ev->y();

  x_down = x, y_down = y;
}

// mouse selection
void GLRecpLattice::mouseDoubleClickEvent( QMouseEvent* ev )
{
}

// mouse drag end
void GLRecpLattice::mouseReleaseEvent( QMouseEvent* ev )
{
}

// mouse drag
void GLRecpLattice::mouseMoveEvent( QMouseEvent* ev )
{
  if( !(ev->buttons()&Qt::LeftButton) ) return;

  int x = ev->x(), y = ev->y();

  const double dx = (1.0/4)*(x-x_down);
  const double dy = (1.0/4)*(y-y_down);

  if( fabs(dx) < 1.0 && fabs(dy) < 1.0 ) return;

  glRotateR( dy, 1.0, 0.0, 0.0 );
  glRotateR( dx, 0.0, 1.0, 0.0 );

  x_down = x, y_down = y;

  updateGL();
}

void GLRecpLattice::wheelEvent( QWheelEvent* ev )
{
  if( ev->orientation() == Qt::Horizontal ) return;

  const double dy = (-2.0) * (double)ev->delta() / 120.0;

  {
    const double f = 1.0 + (1.0/256)*dy;
    glScaled(f,f,f);
  }

  updateGL();
}

void GLRecpLattice::keyPressEvent( QKeyEvent* ev )
{
  double dx=0.0, dy=0.0, dz=0.0;
  switch( ev->key() ){
  case Qt::Key_Left     : dx=-1.0; break;
  case Qt::Key_Right    : dx=+1.0; break;
  case Qt::Key_Up       : dy=-1.0; break;
  case Qt::Key_Down     : dy=+1.0; break;
  case Qt::Key_PageDown : dz=-1.0; break;
  case Qt::Key_PageUp   : dz=+1.0; break;
  default : break;
  }

  if( ev->modifiers() & Qt::ControlModifier ) {
    glRotateR( 15*dy, 1.0, 0.0, 0.0 );
    glRotateR( 15*dx, 0.0, 1.0, 0.0 );
    glRotateR( 15*dz, 0.0, 0.0, 1.0 );
  }
  else{
    switch( ev->key() ){
    case Qt::Key_PageDown :  {
      const double f = 1.0 + (1.0/8);
      glScaled(f,f,f);
    } break; 
    case Qt::Key_PageUp :  {
      const double f = 1.0 - (1.0/8);
      glScaled(f,f,f);
    } break;
    case Qt::Key_Down     : {
      const double dy=-15.0;
      glTranslateR( 0.0, (1.0/4)*(dy), 0.0 );
    } break;
    case Qt::Key_Up       : {
      const double dy=+15.0;
      glTranslateR( 0.0, (1.0/4)*(dy), 0.0 );
    } break;
    case Qt::Key_Left     : {
      const double dx=-15.0;
      glTranslateR( (1.0/4)*(dx), 0.0, 0.0 );
    } break;
    case Qt::Key_Right    : {
      const double dx=+15.0;
      glTranslateR( (1.0/4)*(dx), 0.0, 0.0 );
    } break;

    default : break;
    }
  }

  updateGL();
}


void GLRecpLattice::keyReleaseEvent( QKeyEvent* ev )
{
  if( ev->isAutoRepeat() ) return;

  switch( ev->key() ){
  case Qt::Key_Left   : break;
  case Qt::Key_Right  : break;
  case Qt::Key_Down   : break;
  case Qt::Key_Up     : break;
  case Qt::Key_PageDown : break;
  case Qt::Key_PageUp   : break;
  default             : return;
  }
}

void GLRecpLattice::contextMenuEvent( QContextMenuEvent* ev )
{
  popupmenu->exec(ev->globalPos());
}

void GLRecpLattice::menuEvent( QAction* action )
{
  static bool first = true;
  if( first ){
    path = QDir::currentPath();
    first = false;
  }

  if( false ){
  }
  else if( action->text() == "reset to xy-plane" ||
	   action->text() == "reset to yz-plane" ||
	   action->text() == "reset to zx-plane" ){
    glLoadIdentity();
    gluLookAt( model.gloption.location.eye,
	       model.gloption.location.gaze,
	       model.gloption.location.up );
    model.gloption.light.direction =
      model.gloption.light.direction_orig;
    glLightdv( GL_LIGHT0, GL_POSITION,
	       model.gloption.light.direction );

    if( action->text() == "reset to xy-plane" ){
    }
    if( action->text() == "reset to yz-plane" ){
      glRotateR( -90.0, 0.0, 0.0, 1.0 );
      glRotateR( -90.0, 1.0, 0.0, 0.0 );
    }
    if( action->text() == "reset to zx-plane" ){
      glRotateR( 90.0, 0.0, 1.0, 0.0 );
      glRotateR( 90.0, 1.0, 0.0, 0.0 );
    }
  }
  else if( action->text() == "change perspective/orthogonal" ){
    perspective = !perspective;
    resizeGL( glGetWindowWidth(), glGetWindowHeight() );
  }
  else if( action->text() == "save image" ){
    const QString fname = QFileDialog::getSaveFileName
      ( this, tr("Save image in file"), path,
	tr("Windows Bitmap file(*.bmp);;"
	   "All files(*)"));
    if( fname.endsWith( ".bmp" ) ){
      if( glSavePixels( qPrintable(fname) ) ){
	path = getDirName(fname);
      }
      else{
	MyException::critical("can not create a file.",fname);
      }
    }
    else if( fname !="" ){
      MyException::critical("unknown suffix.",fname);
    }
  }
}



void GLRecpLattice::update( void )
{
  makeCurrent();
  updateGL();
}

void GLRecpLattice::draw( void )
{
  const Position& Ka = model.lattice.cell.Ka;
  const Position& Kb = model.lattice.cell.Kb;
  const Position& Kc = model.lattice.cell.Kc;
  //  const Position& Ka_conv = model.lattice.cell.Ka_conv;
  //  const Position& Kb_conv = model.lattice.cell.Kb_conv;
  //  const Position& Kc_conv = model.lattice.cell.Kc_conv;

  const vector<BrillouinPoint>&   vsampK = model.lattice.cell.vsampK;
  const vector<BrillouinPoint>&   vsymmK = model.lattice.cell.vsymmK;
  const vector<BrillouinSegment>& vsymmL = model.lattice.cell.vsymmL;
  const vector<BrillouinFacet>&   vfacet = model.lattice.cell.vfacet;
  GLcolor color = model.gloption.lattice.cell.color;
  //  const QFont font("Helvetica", 12);
  const QFont font("Currier", 12);

  glDisable( GL_LIGHTING );
  glDisable( GL_DEPTH_TEST );

  glPushMatrix();

  const double L = model.lattice.getReciprocalScale();
  const double f = model.gloption.location.world/L*0.5;
  glScaled(f,f,f);

  glLineWidth(1.0);
  glColor4dv( GLcolor::cyan );

  // primitive k-axes
  glBegin( GL_LINES );
  {
    glVertex3dv( Position(0.0,0.0,0.0) );
    glVertex3dv( Ka );
    glVertex3dv( Position(0.0,0.0,0.0) );
    glVertex3dv( Kb );
    glVertex3dv( Position(0.0,0.0,0.0) );
    glVertex3dv( Kc );
  }
  glEnd();

  {
    this->renderText( Ka.x, Ka.y, Ka.z, QString("Ka"), font );
    this->renderText( Kb.x, Kb.y, Kb.z, QString("Kb"), font );
    this->renderText( Kc.x, Kc.y, Kc.z, QString("Kc"), font );
  }

  // sampling points
  glColor4dv( GLcolor::cyan );
  glPointSize(4.0);
  glBegin( GL_POINTS );
  for( int p=0; p<(int)vsampK.size(); p++ ){
    Position k =
      + Ka * vsampK[p].position.a
      + Kb * vsampK[p].position.b 
      + Kc * vsampK[p].position.c;
    glVertex3dv( k );
  }
  glEnd();
  glPointSize(1.0);

  // symmetric points
  glColor4dv( GLcolor::red );
  glPointSize(4.0);
  glBegin( GL_POINTS );
  for( int p=0; p<(int)vsymmK.size(); p++ ){
    Position k =
      + Ka * vsymmK[p].position.a
      + Kb * vsymmK[p].position.b 
      + Kc * vsymmK[p].position.c;
    glVertex3dv( k );
  }
  glEnd();
  glPointSize(1.0);

  // symmetric lines
  glColor4dv( GLcolor::orange );
  glLineWidth(2.0);
  for( int s=0; s<(int)vsymmL.size(); s++ ){
    Position ks =
      + Ka * vsymmL[s].positions.a
      + Kb * vsymmL[s].positions.b 
      + Kc * vsymmL[s].positions.c;
    Position ke =
      + Ka * vsymmL[s].positione.a
      + Kb * vsymmL[s].positione.b 
      + Kc * vsymmL[s].positione.c;
    glBegin( GL_LINES );
    glVertex3dv( ks );
    glVertex3dv( ke );
    glEnd();
  }
  glLineWidth(1.0);

  // facet
  glColor4dv( color );
  for( int f=0; f<(int)vfacet.size(); f++ ){
    glBegin( GL_LINE_LOOP );
    for( int s=0; s<(int)vfacet[f].vsegment.size(); s++ ){
      glVertex3dv( vfacet[f].vsegment[s].positions );
    }
    glEnd();
  }

  glColor4dv( color );

  // labels of symmetric lines
  for( int s=0; s<(int)vsymmL.size(); s++ ){
    Position ks =
      + Ka * vsymmL[s].positions.a
      + Kb * vsymmL[s].positions.b 
      + Kc * vsymmL[s].positions.c;
    Position ke =
      + Ka * vsymmL[s].positione.a
      + Kb * vsymmL[s].positione.b 
      + Kc * vsymmL[s].positione.c;
    Position kc = (ks+ke)*0.5;

    this->renderText( ks.x, ks.y, ks.z, vsymmL[s].labelPs, font );
    this->renderText( kc.x, kc.y, kc.z, vsymmL[s].labelL , font  );

    if( (s+1==(int)vsymmL.size()) ||
	(s+1< (int)vsymmL.size() &&
	 !Position::match( vsymmL[s].positione, vsymmL[s+1].positions )) ){


      this->renderText( ke.x, ke.y, ke.z, vsymmL[s].labelPe, font );
    }
  }


  glPopMatrix();

  {
    const int win_width  = glGetWindowWidth();
    const int win_height = glGetWindowHeight();

    int xo = 16;
    int yo = 16;

    glColor4dv( GLcolor::white );

    glMatrixMode( GL_MODELVIEW );
    glPushMatrix();
    glLoadIdentity();
  
    glMatrixMode( GL_PROJECTION );
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D( 0.0, (GLfloat)win_width, (GLfloat)win_height, 0.0 );

    char label[256];
    sprintf(label,"Bravais lattice: %s %s %s",
	    qPrintable(model.lattice.cell.bravais.shape),
	    qPrintable(model.lattice.cell.bravais.center),
	    qPrintable(model.lattice.cell.bravais.subtype) );

    this->renderText( xo, yo, QString(label), font );

    glPopMatrix();
    glMatrixMode( GL_MODELVIEW );
    glPopMatrix();
  }

  glEnable( GL_DEPTH_TEST );
  glEnable( GL_LIGHTING );
}
